<?php

$GLOBALS['agi_debug_logging_on'] = true;
$GLOBALS['agi_debug_logging_filename'] = ''; //agi logging file name (full path and file name preferred)

class speech
{
	private $mrcp_profile; // The profile in MRCP.conf to use; used in all interaction types.
	
	// These values are available in mrcp_recog and mrcp_synth_and_recog:
	private $confidence_threshold; // The confidence threshold for a recognition, 0.0-1.0
	private $barge_in; // 0 = no barge-in, 1 = ASR barge-in, 3 = Asterisk barge-in. Use 1.
	private $sensitivity; // Barge-in sensitivity, higher is easier to barge-in. Default 0.5; 0.0-1.0
	private $speed_vs_accuracy; // Higher is faster but less accurate. Default 0.5; 0.0-1.0.
	private $max_nbest; // The maximum number of n-best results. Default is 1.
	private $recognition_timeout; // Total amount of time caller has to speak, in ms.
	private $no_input_timeout; // Length of time caller has to start speaking, in ms.
	private $speech_complete_timeout; // Length of time caller has to pause after speaking, in ms.
	private $dtmf_interdigit_timeout; // DTMF inter-digit timeout, in ms.
	private $dtmf_terminate_timeout; // DTMF termination timeout, in ms.
	
	// These values available in mrcp_synth and mrcp_synth_and_recog:
	private $synth_language; // Default language to use, e.g. en-US or es-MX.
	private $voice_name; // Voice name to use, e.g. Lindsey
	private $voice_gender; // Gender gender, male or female
	private $prosody_volume; // Volume of synthesis, silent/x-soft/soft/medium/loud/x-loud/default
	private $prosody_rate; // Rate of speech, x-slow/slow/medium/fast/x-fast/default		
	
	private $option_string; // The option string that gets passed into the MRCP application.
	
	private $no_match_text = "I am sorry, but I did not understand what you said";
	private $no_input_text = "I am sorry, but I did not hear what you said";
	
	// Performs TTS and ASR at the same time, taking in optional arguments.
	// Returns an object of the class speech_return.
	// We *strongly* recommend passing in URIs to SSML and grammar files and not 
	// attempting to add them in the contents as text strings as this will lead
	// to all kinds of weird behavior. Also note that you should not pass in commas
	// as those are reserved in dialplan. The function will escape spaces and strip out
	// commas, but otherwise it does little sanity checking on inputs. Be warned.
	public function mrcp_synth_and_recog($text, $grammar){
		$text = str_replace(" ", "\ ", $text);
		$text = str_replace(",", "", $text);
		$exec_string = 'EXEC SynthAndRecog ';
		$exec_string .= $text;
		$exec_string .= ',' . $grammar;
		$interaction_type = "synth_and_recog";
		// Convert the set options into a string that the dialplan application can understand
		$this->option_string = $this->set_option_string($interaction_type);

		if ($this->option_string != "" ){
			$exec_string .= ',' . $this->option_string;
		}
		execute_agi($exec_string);
		return $this->get_speech_return($interaction_type);		
	}
	
	// Performs ASR while playing out an optional audio file.
	public function mrcp_recog($grammar, $audio = ""){
		error_log('In mrcp_recog');
		
		$exec_string = 'EXEC MRCPRecog ';
		$exec_string .= $grammar . ",f=" . $audio;
		$interaction_type = "recog";
		$this->option_string .= $this->set_option_string($interaction_type);
		if ($this->option_string != "" ){
			$exec_string .= "&" . $this->option_string;
		}
		dump_var_to_error($exec_string);
		// Calls the recog dialplan app
		execute_agi($exec_string);
		// Get the results and pass it out
		return $this->get_speech_return($interaction_type);	
	}

	// Performs a synthesis on some text and takes some optional parameters.
	public function mrcp_synth($text){

		$text = str_replace(" ", "\ ", $text);
		$text = str_replace(",", "", $text);
		$text = str_replace("'", "\'", $text);
		$exec_string = 'EXEC MRCPSynth ';
		$exec_string .= $text;
		$interaction_type = "synth";
		$this->option_string = $this->set_option_string($interaction_type);	
		if ($this->option_string != "" ){
			$exec_string .= ',' . $this->option_string;
		}		
		// Calls the synthesis
		execute_agi($exec_string);		
		return get_dialplan_variable_data('SYNTHSTATUS');		
	}
	
	// Creates a new speech_return object and populates it with information from the last recognition
	private function get_speech_return($interaction_type){
		$new_speech_return = new speech_return;
		$new_speech_return->recognition_result = get_dialplan_variable_data('RECOG_COMPLETION_CAUSE');
		
		// Insanely, mrcp_synth_and_recog and mrcp_recog call the status return something different
		if ($interaction_type == "synth_and_recog"){
			$new_speech_return->recognition_status = get_dialplan_variable_data('RECOG_STATUS');
		}
		else {
			$new_speech_return->recognition_status = get_dialplan_variable_data('RECOGSTATUS');
		}
		// If there was a result, we can get the NLSML, parse it, and build our answers. If 
		// there's no input, there's really nothing meaningful in the NLSML to parse, so it's a good idea 
		// to not even try.
		if ($new_speech_return->recognition_result==speech_return::result_input){
			$new_speech_return->nlsml_string = get_dialplan_variable_data('RECOG_RESULT');
			$new_speech_return->answers = $this->parse_nlsml($new_speech_return);
		}
		return $new_speech_return;
	}

	// Parses an NLSML-formatted XML string, and puts data from it into a speech_return object.
	private function parse_nlsml($new_speech_return){
		// There are two broad reasons you might be unable to parse NLSML: either there was no 
		// result (get_speech_return() should not attempt the parse in this case) 
		// or for some reason the NLSML is invalid. This can happen in Asterisk with very long 
		// NLSML results due to limitations in Asterisk. If this happens, we'll return an empty results 
		// array. 
		try {
			$result = new SimpleXMLElement($new_speech_return->nlsml_string);
		}				
		catch (Exception $e) {
			log_agi('Unable to parse NSML');
			$result = array();
			return $result;
		}

		// Set the input mode (either speech or DTMF) based on the NLSML. Theoretically 
		// NLSML can return mixed input modes in different answers, but LumenVox does not 
		// support this behavior, so we can simply get the first mode value and use it for 
		// the entire speech_return		
		$new_speech_return->mode=$result->interpretation[0]->input->attributes()->mode;	

		/*	SimpleXMLElement already parsed the NLSML, but the structure of NLSML is 
			convoluted, so we are going to return a differently-structured object that is easier
			to work with. We do this by iterating on the "interpretation" node in the NLSML object
			and building up an array called "answers". The basic structure is for answers to be an 
			enumerated array, with each element in the array an object of the speech_answer type. 
			Any given answer may have one or more interpretations associated with it; interpretations 
			is also an array. The structure of any given interpretation element is defined by the 
			speech grammar, however if it is simply a string we'll make sure that the first (and only) 
			element in the interpretation array is a string.
		
			As an example, a simple 1-answer, 1-interpretation return might look like:
			
			answers
				[0]
					confidence : 50
					grammar : yesno.gram
					input : yes please
					interpretation
						[0] => yes
		
			We might also have a more complex result, with 2 n-best answers and a varying number of interpretations:
			
			answers
				[0]
					confidence : 75
					grammar : names.gram
					input : john
					interpretation
						[0]
							firstName
								John
							lastName
								Smith
						[1]
							firstName
								John
							lastName
								Jones
				[1]
					confidence : 20
					grammar : names.gram
					input : juan
					interpretation
						[0]
							firstName
								Juan
							lastName
								Smith
			
			In most cases, developers can simply check answers[0]->interpretation[0] and get the meaning.
			The number of answers or interpretations at any level can also be easily checked with sizeof(answers[0]) 
			or sizeof(interpretation[0]).
			
			For more about the various ways NLSML results get formatted, please read the LumenVox documentation
			on NLSML: http://www.lumenvox.com/knowledgebase/index.php?/article/AA-01803/0/						
					
		*/
		$answer_index = 0;	
		foreach ($result->interpretation as $value) {
			
			$answers[$answer_index] = new speech_answer;
			$answers[$answer_index]->confidence = intval($value->attributes()->confidence);	
			$answers[$answer_index]->grammar = strval(trim($value->attributes()->grammar));
			$answers[$answer_index]->input = strval(trim($value->input[0]));

			$interp_index = 0;
			foreach($value->instance as $interp) {
				if (count($interp) == 0) {
					$answers[$answer_index]->interpretation[$interp_index] = strval($interp);					
				}
				else {
					$answers[$answer_index]->interpretation[$interp_index] = $interp;
				}
				$interp_index++;		
			}
			$answer_index++;
		}
		return $answers;	
	}

	// A simple function that returns true if there was a match and false; 
	// also calls the appropriate no_input and no_match functions in those cases
	public function check_match($speech_return) {
		switch ($speech_return->recognition_result){
			case speech_return::result_no_input:
				$this->no_input();
				$match_exists = false;
				break;
			case speech_return::result_no_match:
				$this->no_match();
				$match_exists = false;
				break;
			case speech_return::result_timeout:
				$this->no_match();
				$match_exists = false;
				break;
			case speech_return::result_input:
				$match_exists = true;
				break;
		}
		return $match_exists;
	}
	
	// A simple function to check whether a speech_return has any ambiguity; i.e. 
	// whether there are multiple answers or, if only one, are there multiple interpretations. 
	// We don't need to handle the case where there are more multiple answers and multiple 
	// interpretations since that's implicitly taken care of in the first case.
	public function check_ambiguity($speech_return) {
		if (sizeof($speech_return->answers) > 1){
			$disambig_needed = true;
			return $disambig_needed;
		}
		if (sizeof($speech_return->answers[0]->interpretation) > 1) {
			$disambig_needed = true;
		}
		else {
			$disambig_needed = false;
		}		
		return $disambig_needed;
	}
	
	// Asks a yes or no question. Loops until there is a clear answer.
	public function ask_yes_or_no($question){
		$grammar = "builtin:grammar/boolean";
		$yes_no_done = false;
		$yes_no_result = "";
		while(!$yes_no_done){
			$yes_no_answer = $this->mrcp_synth_and_recog($question, $grammar);
			if ($yes_no_answer->recognition_status!=speech_return::status_successful){
				// In a real application, this is where you might transfer to an operator or revert to DTMF.
				log_agi('There was an error with speech recognition and/or TTS.');
				exit();
			}
			// Did we get at least one match?
			if (!$this->check_match($yes_no_answer)) {
				$yes_no_done = false;
			}
			else {
				$yes_no_done = true;	
				if($yes_no_answer->answers[0]->interpretation[0] == "true"){
						$yes_no = true;
					}
				else{
					$yes_no = false;
				}
			}
		}
		return $yes_no;
	}	

	// Plays a no input message. This can (and should) be customized.
	public function no_input() {
		$this->mrcp_synth($this->no_input_text);
	}

	// Plays a no match message. This can (and should) be customized.
	function no_match() {	
		$this->mrcp_synth($this->no_match_text);
	}
	
	// Set and Get methods for all options:
	
	public function set_mrcp_profile($value){
		if (!is_string($value)){
			throw new Exception("Profile name must be a string.");
		}	
		$this->mrcp_profile =  $value;
	}	
	
	public function get_mrcp_profile(){
		return $this->mrcp_profile;
	}
	
	public function set_confidence_threshold($value){
		if (!is_float($value)){
			throw new Exception("Confidence value must be a float.");
		}
		if ($value < 0.0 || $value > 1.0){
			throw new Exception("Confidence value out of range.");
		}
		$this->confidence_threshold =  $value;
	}

	public function get_confidence_threshold(){
		return $this->confidence_threshold;
	}		

	public function set_barge_in($value){
		if (!is_int($value)){
			throw new Exception("Barge-in value must be an int.");
		}
		if ($value < 0 || $value > 2){
			throw new Exception("Barge-in value out of range.");
		}
		$this->barge_in =  $value;
	}

	public function get_barge_in(){
		return $this->barge_in;
	}
		
	
	public function set_sensitivity($value){
		if (!is_float($value)){
			throw new Exception("Sensitivity value must be a float.");
		}
		if ($value < 0.0 || $value > 1.0){
			throw new Exception("Sensitivity value out of range.");
		}
		$this->sensitivity =  $value;
	}

	public function get_sensitivity(){
		return $this->sensitivity;
	}
	
	public function set_speed_vs_accuracy($value){
		if (!is_float($value)){
			throw new Exception("Speed vs Accuracy value must be a float.");
		}
		if ($value < 0.0 || $value > 1.0){
			throw new Exception("Speed vs Accuracy value out of range.");
		}
		$this->speed_vs_accuracy =  $value;
	}

	public function get_speed_vs_accuracy(){
		return $this->speed_vs_accuracy;
	}
	
	public function set_max_nbest($value){
		if (!is_int($value)){
			throw new Exception("Max N-best value must be an int.");
		}
		if ($value < 1){
			throw new Exception("Max N-best value out of range.");
		}
		$this->max_nbest =  $value;
	}

	public function get_max_nbest(){
		return $this->max_nbest;
	}

	public function set_recognition_timeout($value){
		if (!is_int($value)){
			throw new Exception("Recognition timeout value must be an int.");
		}
		if ($value < 1){
			throw new Exception("Recognition timeout value out of range.");
		}
		$this->recognition_timeout =  $value;
	}

	public function get_recognition_timeout(){
		return $this->recognition_timeout;
	}
	
	public function set_no_input_timeout($value){
		if (!is_int($value)){
			throw new Exception("No input timeout value must be an int.");
		}
		if ($value < 1){
			throw new Exception("No input timeout value out of range.");
		}
		$this->no_input_timeout =  $value;
	}

	public function get_no_input_timeout(){
		return $this->no_input_timeout;
	}
	
	public function set_speech_complete_timeout($value){
		if (!is_int($value)){
			throw new Exception("Speech complete timeout value must be an int.");
		}
		if ($value < 1){
			throw new Exception("Speech complete timeout value out of range.");
		}
		$this->speech_complete_timeout =  $value;
	}

	public function get_speech_complete_timeout(){
		return $this->speech_complete_timeout;
	}
	
	public function set_dtmf_interdigit_timeout($value){
		if (!is_int($value)){
			throw new Exception("DTMF interdigit timeout value must be an int.");
		}
		if ($value < 1){
			throw new Exception("DTMF interdigit timeout value out of range.");
		}
		$this->dtmf_interdigit_timeout =  $value;
	}

	public function get_dtmf_interdigit_timeout(){
		return $this->dtmf_interdigit_timeout;
	}
	
	public function set_dtmf_terminate_timeout($value){
		if (!is_int($value)){
			throw new Exception("DTMF termination timeout value must be an int.");
		}
		if ($value < 1){
			throw new Exception("DTMF termination timeout value out of range.");
		}
		$this->dtmf_terminate_timeout =  $value;
	}

	public function get_dtmf_terminate_timeout(){
		return $this->dtmf_terminate_timeout;
	}

	public function set_synth_language($value){
		if (!is_string($value)){
			throw new Exception("Synth language must be a string.");
		}
		$this->synth_language =  $value;
	}

	public function get_synth_language(){
		return $this->synth_language;
	}
	
	public function set_voice_name($value){
		if (!is_string($value)){
			throw new Exception("Voice name must be a string.");
		}
		$this->voice_name =  $value;
	}

	public function get_voice_name(){
		return $this->voice_name;
	}
	
	public function set_voice_gender($value){
		if (!is_string($value)){
			throw new Exception("Voice gender must be a string.");
		}
		$this->voice_gender =  $value;
	}

	public function get_voice_gender(){
		return $this->voice_gender;
	}
	
	public function set_prosody_volume($value){
		if (!is_string($value)){
			throw new Exception("Prosody volume must be a string.");
		}
		$this->prosody_volume =  $value;
	}

	public function get_prosody_volume(){
		return $this->prosody_volume;
	}

	public function set_prosody_rate($value){
		if (!is_string($value)){
			throw new Exception("Prosody rate must be a string.");
		}
		$this->prosody_rate =  $value;
	}

	public function get_prosody_rate(){
		return $this->prosody_rate;
	}

	public function set_no_match_text($value){
		if (!is_string($value)){
			throw new Exception("No match text  must be a string.");
		}
		$this->no_match_text = $value;
	}

	public function get_no_match_text(){
		return $this->no_match_text;
	}

	public function set_no_input_text($value){
		if (!is_string($value)){
			throw new Exception("No input text  must be a string.");
		}
		$this->no_match_text = $value;
	}

	public function get_no_input_text(){
		return $this->no_input_text;
	}
	
	
	// Turns the options into a "parameter=value,&parameter=value" string for the dialplan applications
	private function set_option_string($interaction_type){
		$option_array = array();
		$this->build_option_array("mrcp_profile","p",$option_array);
				
		if ($interaction_type == "recog" || $interaction_type == "synth_and_recog"){$this->build_option_array("confidence_threshold","ct",$option_array);}
		if ($interaction_type == "recog" || $interaction_type == "synth_and_recog"){$this->build_option_array("barge_in","b",$option_array);}		
		if ($interaction_type == "recog" || $interaction_type == "synth_and_recog"){$this->build_option_array("sensitivity","sl",$option_array);}
		if ($interaction_type == "recog" || $interaction_type == "synth_and_recog"){$this->build_option_array("speed_vs_accuracy","sva",$option_array);}
		if ($interaction_type == "recog" || $interaction_type == "synth_and_recog"){$this->build_option_array("max_nbest","nb",$option_array);}
		if ($interaction_type == "recog" || $interaction_type == "synth_and_recog"){$this->build_option_array("recognition_timeout","t",$option_array);}		
		if ($interaction_type == "recog" || $interaction_type == "synth_and_recog"){$this->build_option_array("no_input_timeout","nit",$option_array);}
		if ($interaction_type == "recog" || $interaction_type == "synth_and_recog"){$this->build_option_array("speech_complete_timeout","sct",$option_array);}
		if ($interaction_type == "recog" || $interaction_type == "synth_and_recog"){$this->build_option_array("dtmf_interdigit_timeout","dit",$option_array);}
		if ($interaction_type == "recog" || $interaction_type == "synth_and_recog"){$this->build_option_array("dtmf_terminate_timeout","dtt",$option_array);}
		
		if ($interaction_type == "synth" || $interaction_type == "synth_and_recog"){$this->build_option_array("prosody_volume","pv",$option_array);}
		if ($interaction_type == "synth" || $interaction_type == "synth_and_recog"){$this->build_option_array("prosody_rate","pr",$option_array);}

		if ($interaction_type == "synth_and_recog"){$this->build_option_array("voice_name","vn",$option_array);}
		if ($interaction_type == "synth_and_recog"){$this->build_option_array("voice_gender","vg",$option_array);}		
		
		if ($interaction_type == "synth"){$this->build_option_array("voice_name","v",$option_array);}
		if ($interaction_type == "synth"){$this->build_option_array("voice_gender","g",$option_array);}					
		if ($interaction_type == "synth"){$this->build_option_array("synth_language","l",$option_array);}
		$option_string = implode("&", $option_array);
		return $option_string;
	}
	
	private function build_option_array($option,$mrcp_parameter,&$option_array){
		if(isset($this->{$option})){
			$option_array[] = $mrcp_parameter . "=" . strval($this->{$option});
		}
	}
}

// A return from a recognition event.
class speech_return 
{
	// Define results from UniMRCP to indicate various possibilites when performing speech
	const result_input = '000';
	const result_no_match  = '001';
	const result_no_input   = '002';	
	const result_timeout  = '003';
	
	// Define statuses that MRCP dialplan applications return
	const status_successful = 'OK';
	const status_error = 'ERROR';
	const status_interrupted = 'INTERRUPTED';
	
	public $recognition_status = "";
	public $recognition_result = "";
	public $nlsml_string = "";
	public $mode = "";
	public $answers = array();
}

// An individual answer
class speech_answer 
{	
	public $confidence = 0;
	public $grammar = "";
	public $input = "";
	public $interpretation=array();
}

// Get a dialplan variable, including return codes
function get_dialplan_variable($var_name) {
	return execute_agi('GET VARIABLE ' . $var_name);
}


// Get just the data related to a dialplan variable. Useful when you don't care about the various codes.
function get_dialplan_variable_data($var_name) {
	$result = get_dialplan_variable($var_name);
	$ret = $result->extra;
	return $ret;
}


// Executes an AGI command and parses the output
function execute_agi($command) 
{
	$agi_debug_logging_on = false;
	$agi_debug_logging_filename = '';
	
	if (isset($GLOBALS['agi_debug_logging_on'])) $agi_debug_logging_on = $GLOBALS['agi_debug_logging_on'];
	if (isset($GLOBALS['agi_debug_logging_filename'])) $agi_debug_logging_filename = $GLOBALS['agi_debug_logging_filename'];

	fwrite(STDOUT, "$command\n");
	fflush(STDOUT);
	$result = trim(fgets(STDIN));
	$return = new stdclass();  
	$return->code = -1;  $return->result = -1;  $return->timeout = false;  $return->extra = '';

	if (preg_match("/^([0-9]{1,3}) (.*)/", $result, $matches)) 
		{
		 $return->code = $matches[1];
		 $return->result = 0;
		 if (preg_match('/^result=([0-9a-zA-Z]*)\s?(?:\(?(.*?)\)?)?$/', $matches[2], $match))  
			{
			 $return->result = $match[1];
			 $return->extra = $match[2];
			 if ($return->extra === 'timeout') $return->timeout = true;
			}
		}
		
	if ($agi_debug_logging_on && $agi_debug_logging_filename != '') 
		{
		 $msg = "Cmd: $command  Code: $return->code  Result: $return->result";
		 if (!empty($return->extra)) $msg .= "  Data: $return->extra";
 		 error_log(date("m-d H:i:s ") . $msg . "\n", 3, $agi_debug_logging_filename);
		}
	
	return $return;
}

// log out to the Asterisk console
function debug_log_agi($message, $verbosity = 1) 
{
     if (!is_numeric($verbosity)) $verbosity = 1;
     if ($verbosity < 1) $verbosity = 1;
     if ($verbosity > 4) $verbosity = 4;
     $result = execute_agi("VERBOSE" . ' "' . $message . '" ' . $verbosity);
}

 
function get_agi_variables()
{
	$agi_variables = array();
	while (!feof(STDIN)) 
		{
		 $variable = trim(fgets(STDIN));
		 if ($variable === '') 
			 break;
		else
			{
			 $pair = explode(':', $variable);
			 $agi_variables[$pair[0]] = trim($pair[1]);
			}
		}
	return $agi_variables;
}


?>